"""Makes sure static context manager works as expected."""

import pytest
import time

import numpy as np

import tidy3d as td
from tidy3d.static import MakeStatic, StaticException, make_static


def make_sim(num_structures=100):
    """Returns a fresh simulation."""
    return td.Simulation(
        size=(5, 5, 5),
        grid_spec=td.GridSpec(wavelength=1.0),
        run_time=1e-12,
        structures=[
            td.Structure(
                geometry=td.Box(size=(1, 1, 1)),
                medium=td.Medium(permittivity=1 + np.random.random()),
            )
            for i in range(num_structures)
        ],
        sources=[
            td.PointDipole(
                center=(0, 0.5, 0),
                polarization="Ex",
                source_time=td.GaussianPulse(
                    freq0=2e14,
                    fwidth=4e13,
                ),
            )
        ],
    )


def test_context_manager_works_static():
    """Makes sure the context manager works as expected when sim not mutated."""

    sim_original = make_sim()

    assert sim_original._hash is None, "sim hash is not None to start."

    with MakeStatic(sim_original) as sim_static:

        assert sim_static._hash is not None, "static sim._hash should not be None"

        run_time_x2 = sim_original.run_time * 2

    assert sim_original._hash is None


def test_context_manager_fails_on_mutation():
    """Makes sure the context manager raises StaticException when sim_static changed."""

    sim_original = make_sim()

    assert sim_original._hash is None, "sim hash is not None to start."

    with pytest.raises(StaticException):
        with MakeStatic(sim_original) as sim_static:
            sim_static.run_time *= 2

    assert sim_original._hash is None


def test_context_manager_fails_on_list():
    """Makes sure the context manager raises StaticException when sim_static's attr mutates."""

    sim_original = make_sim()

    assert sim_original._hash is None, "sim hash is not None to start."

    with pytest.raises(StaticException):
        with MakeStatic(sim_original) as sim_static:
            sim_static.sources[0].source_time.freq0 *= 2

    assert sim_original._hash is None


def test_decorator_works_static():
    """Make sure decorator works in static case, with many args."""

    @make_static
    def mult_and_shift_run_time(simulation, mult_factor, shift_factor=1e-12) -> float:
        """Multiply run time by factor and shift by another factor."""
        return simulation.run_time * mult_factor + shift_factor

    sim_original = make_sim()

    ret = mult_and_shift_run_time(sim_original, 2)


def test_decorator_fails_mutate():
    """Make sure decorator fails in mutated case, with many args."""

    @make_static
    def mult_and_shift_run_time_inplace(simulation, mult_factor, shift_factor=1e-12) -> None:
        """Augment simulation run time by multiplicative factor and shift by another factor."""
        simulation.run_time *= mult_factor
        simulation.run_time += shift_factor

    sim_original = make_sim()

    with pytest.raises(StaticException):
        ret = mult_and_shift_run_time_inplace(sim_original, 2)


def test_one_arg_static():
    """Make sure decorator works in static case when only one arg supplied."""

    @make_static
    def run_time_squared(simulation) -> float:
        """Return 2x simulation run_time."""
        return simulation.run_time * 2

    sim_original = make_sim()

    ret = run_time_squared(sim_original)


def test_one_arg_static():
    """Make sure decorator works in mutated case when only one arg supplied."""

    @make_static
    def run_time_squared_inplace(simulation) -> float:
        """double simulation run time."""
        simulation.run_time *= 2

    sim_original = make_sim()

    with pytest.raises(StaticException):
        ret = run_time_squared_inplace(sim_original)


def sum_of_N_timesteps(sim, N=50_000):
    """sum of len(dt) computed N times."""
    ret = 0
    for _ in range(N):
        ret += len(sim.tmesh)
    return ret


@pytest.mark.timeout(3.0)
def test_grid_run_fast():
    """Make sure repeated .grid computation is fast."""
    sim = make_sim()
    sum_of_N_timesteps_static = make_static(sum_of_N_timesteps)
    sum = sum_of_N_timesteps_static(sim, N=10_000)



"""Near field to far field transformation plugin
"""
from typing import List, Dict, Tuple, Union
import numpy as np
import xarray as xr
import pydantic

from rich.progress import track

from ...constants import C_0, ETA_0, HERTZ, MICROMETER
from ...components.data import SimulationData, FieldData
from ...components.monitor import FieldMonitor
from ...components.types import Direction, Axis, Coordinate, ArrayLike
from ...components.medium import Medium
from ...components.base import Tidy3dBaseModel
from ...log import SetupError, ValidationError

# Default number of points per wavelength in the background medium to use for resampling fields.
PTS_PER_WVL = 10

# Numpy float array and related array types
ArrayLikeN2F = Union[float, List[float], ArrayLike]


class Near2FarSurface(Tidy3dBaseModel):
    """Data structure to store surface monitor data with associated surface current densities."""

    monitor: FieldMonitor = pydantic.Field(
        ...,
        title="Field monitor",
        description=":class:`.FieldMonitor` on which near fields will be sampled and integrated.",
    )

    normal_dir: Direction = pydantic.Field(
        ...,
        title="Normal vector orientation",
        description=":class:`.Direction` of the surface monitor's normal vector w.r.t.\
 the positive x, y or z unit vectors. Must be one of '+' or '-'.",
    )

    @property
    def axis(self) -> Axis:
        """Returns the :class:`.Axis` normal to this surface."""
        # assume that the monitor's axis is in the direction where the monitor is thinnest
        return self.monitor.size.index(0.0)

    @pydantic.validator("monitor", always=True)
    def is_plane(cls, val):
        """Ensures that the monitor is a plane, i.e., its `size` attribute has exactly 1 zero"""
        size = val.size
        if size.count(0.0) != 1:
            raise ValidationError(f"Monitor '{val.name}' must be planar, given size={size}")
        return val


class Near2Far(Tidy3dBaseModel):
    """Near field to far field transformation tool."""

    sim_data: SimulationData = pydantic.Field(
        ...,
        title="Simulation data",
        description="Container for simulation data containing the near field monitors.",
    )

    surfaces: List[Near2FarSurface] = pydantic.Field(
        ...,
        title="Surface monitor with direction",
        description="List of each :class:`.Near2FarSurface` to use as source of near field.",
    )

    frequency: float = pydantic.Field(
        ...,
        title="Frequency",
        description="Frequency to select from each :class:`.Near2FarSurface` for projection.",
        units=HERTZ,
    )

    pts_per_wavelength: int = pydantic.Field(
        PTS_PER_WVL,
        title="Points per wavelength",
        description="Number of points per wavelength in the background medium with which "
        "to discretize the surface monitors for the projection.",
    )

    medium: Medium = pydantic.Field(
        None,
        title="Background medium",
        description="Background medium in which to radiate near fields to far fields. "
        "If ``None``, uses the :class:.Simulation background medium.",
    )

    origin: Coordinate = pydantic.Field(
        None,
        title="Local origin",
        description="Local origin used for defining observation points. If ``None``, uses the "
        "average of the centers of all surface monitors.",
        units=MICROMETER,
    )

    currents: Dict[str, xr.Dataset] = pydantic.Field(
        None,
        title="Surface current densities",
        description="Dictionary mapping monitor name to an ``xarray.Dataset`` storing the "
        "surface current densities.",
    )

    phasor_positive_sign: bool = pydantic.Field(
        1,
        title="Phasor convention",
        description="Fields evolve as exp(jkr) if set to 1, and exp(-jkr) if set to -1. "
        "Should not be changed except in special cases where the exp(-jkr) convention is used.",
    )

    @pydantic.validator("origin", always=True)
    def set_origin(cls, val, values):
        """Sets .origin as the average of centers of all surface monitors if not provided."""
        if val is None:
            surfaces = values.get("surfaces")
            val = np.array([surface.monitor.center for surface in surfaces])
            return tuple(np.mean(val, axis=0))
        return val

    @pydantic.validator("medium", always=True)
    def set_medium(cls, val, values):
        """Sets the .medium field using the simulation default if no medium was provided."""
        if val is None:
            val = values.get("sim_data").simulation.medium
        return val

    @property
    def nk(self) -> Tuple[float, float]:
        """Returns the real and imaginary parts of the background medium's refractive index."""
        eps_complex = self.medium.eps_model(self.frequency)
        return self.medium.eps_complex_to_nk(eps_complex)

    @property
    def k(self) -> complex:
        """Returns the complex wave number associated with the background medium."""
        index_n, index_k = self.nk
        return (2 * np.pi * self.frequency / C_0) * (index_n + 1j * index_k)

    @property
    def eta(self) -> complex:
        """Returns the complex wave impedance associated with the background medium."""
        index_n, index_k = self.nk
        return ETA_0 / (index_n + 1j * index_k)

    @classmethod
    # pylint:disable=too-many-arguments
    def from_surface_monitors(
        cls,
        sim_data: SimulationData,
        monitors: List[FieldMonitor],
        normal_dirs: List[Direction],
        frequency: float,
        pts_per_wavelength: int = PTS_PER_WVL,
        medium: Medium = None,
        origin: Coordinate = None,
    ):
        """Constructs :class:`Near2Far` from a list of surface monitors and their directions.

        Parameters
        ----------
        sim_data : :class:`.SimulationData`
            Container for simulation data containing the near field monitors.
        monitors : List[:class:`.FieldMonitor`]
            List of :class:`.FieldMonitor` objects on which near fields will be sampled.
        normal_dirs : List[:class:`.Direction`]
            List containing the :class:`.Direction` of the normal to each surface monitor
            w.r.t. to the positive x, y or z unit vectors. Must have the same length as monitors.
        frequency : float
            Frequency to select from each :class:`.FieldMonitor` to use for projection.
            Must be a frequency stored in each :class:`FieldMonitor`.
        pts_per_wavelength : int = 10
            Number of points per wavelength with which to discretize the
            surface monitors for the projection.
        medium : :class:`.Medium`
            Background medium in which to radiate near fields to far fields.
            Default: same as the :class:`.Simulation` background medium.
        origin : :class:`.Coordinate`
            Local origin used for defining observation points. If ``None``, uses the
            average of the centers of all surface monitors.
        """

        if len(monitors) != len(normal_dirs):
            raise SetupError(
                f"Number of monitors ({len(monitors)}) does not equal \
the number of directions ({len(normal_dirs)})."
            )

        surfaces = []
        for monitor, normal_dir in zip(monitors, normal_dirs):
            surfaces.append(Near2FarSurface(monitor=monitor, normal_dir=normal_dir))

        return cls(
            sim_data=sim_data,
            surfaces=surfaces,
            frequency=frequency,
            pts_per_wavelength=pts_per_wavelength,
            medium=medium,
            origin=origin,
        )

    @pydantic.validator("currents", always=True)
    def set_currents(cls, val, values):
        """Sets the surface currents."""
        sim_data = values.get("sim_data")
        surfaces = values.get("surfaces")
        pts_per_wavelength = values.get("pts_per_wavelength")
        frequency = values.get("frequency")
        medium = values.get("medium")
        eps_complex = medium.eps_model(frequency)
        index_n, _ = medium.eps_complex_to_nk(eps_complex)

        val = {}
        for surface in surfaces:
            current_data = cls.compute_surface_currents(
                sim_data, surface, frequency, index_n, pts_per_wavelength
            )
            val[surface.monitor.name] = current_data

        return val

    @staticmethod
    def compute_surface_currents(
        sim_data: SimulationData,
        surface: Near2FarSurface,
        frequency: float,
        index_n: float,
        pts_per_wavelength: int = PTS_PER_WVL,
    ) -> xr.Dataset:
        """Returns resampled surface current densities associated with the surface monitor.

        Parameters
        ----------
        sim_data : :class:`.SimulationData`
            Container for simulation data containing the near field monitors.
        surface: :class:`.Near2FarSurface`
            :class:`.Near2FarSurface` to use as source of near field.
        frequency : float
            Frequency to select from each :class:`.FieldMonitor` to use for projection.
            Must be a frequency stored in each :class:`FieldMonitor`.
        index_n : float
            Real part of the refractive index associated with the background medium.
        pts_per_wavelength : int = 10
            Number of points per wavelength with which to discretize the
            surface monitors for the projection.

        Returns
        -------
        xarray.Dataset
            Colocated surface current densities for the given surface.
        """

        try:
            field_data = sim_data[surface.monitor.name]
        except Exception as e:
            raise SetupError(
                f"No data for monitor named '{surface.monitor.name}' found in sim_data."
            ) from e

        currents = Near2Far._fields_to_currents(field_data, surface)
        currents = Near2Far._resample_surface_currents(
            currents, sim_data, surface, frequency, index_n, pts_per_wavelength
        )

        return currents

    @staticmethod
    def _fields_to_currents(field_data: FieldData, surface: Near2FarSurface) -> FieldData:
        """Returns surface current densities associated with a given :class:`.FieldData` object.

        Parameters
        ----------
        field_data : :class:`.FieldData`
            Container for field data associated with the given near field surface.
        surface: :class:`.Near2FarSurface`
            :class:`.Near2FarSurface` to use as source of near field.

        Returns
        -------
        :class:`.FieldData`
            Surface current densities for the given surface.
        """

        # figure out which field components are tangential or normal to the monitor
        normal_field, tangent_fields = surface.monitor.pop_axis(("x", "y", "z"), axis=surface.axis)

        signs = np.array([-1, 1])
        if surface.axis % 2 != 0:
            signs *= -1
        if surface.normal_dir == "-":
            signs *= -1

        # compute surface current densities and delete unneeded field components
        currents = field_data.copy(deep=True)
        cmp_1, cmp_2 = tangent_fields

        currents.data_dict["J" + cmp_2] = currents.data_dict.pop("H" + cmp_1)
        currents.data_dict["J" + cmp_1] = currents.data_dict.pop("H" + cmp_2)
        del currents.data_dict["H" + normal_field]

        currents.data_dict["M" + cmp_2] = currents.data_dict.pop("E" + cmp_1)
        currents.data_dict["M" + cmp_1] = currents.data_dict.pop("E" + cmp_2)
        del currents.data_dict["E" + normal_field]

        currents.data_dict["J" + cmp_1].values *= signs[0]
        currents.data_dict["J" + cmp_2].values *= signs[1]

        currents.data_dict["M" + cmp_1].values *= signs[1]
        currents.data_dict["M" + cmp_2].values *= signs[0]

        return currents

    @staticmethod
    # pylint:disable=too-many-locals, too-many-arguments
    def _resample_surface_currents(
        currents: xr.Dataset,
        sim_data: SimulationData,
        surface: Near2FarSurface,
        frequency: float,
        index_n: float,
        pts_per_wavelength: int = PTS_PER_WVL,
    ) -> xr.Dataset:
        """Returns the surface current densities associated with the surface monitor.

        Parameters
        ----------
        currents : xarray.Dataset
            Surface currents defined on the original Yee grid.
        sim_data : :class:`.SimulationData`
            Container for simulation data containing the near field monitors.
        surface: :class:`.Near2FarSurface`
            :class:`.Near2FarSurface` to use as source of near field.
        frequency : float
            Frequency to select from each :class:`.FieldMonitor` to use for projection.
            Must be a frequency stored in each :class:`FieldMonitor`.
        index_n : float
            Real part of the refractive index associated with the background medium.
        pts_per_wavelength : int = 10
            Number of points per wavelength with which to discretize the
            surface monitors for the projection.

        Returns
        -------
        xarray.Dataset
            Colocated surface current densities for the given surface.
        """

        # colocate surface currents on a regular grid of points on the monitor based on wavelength
        colocation_points = [None] * 3
        colocation_points[surface.axis] = surface.monitor.center[surface.axis]

        wavelength = C_0 / frequency / index_n

        _, idx_uv = surface.monitor.pop_axis((0, 1, 2), axis=surface.axis)

        for idx in idx_uv:

            # pick sample points on the monitor and handle the possibility of an "infinite" monitor
            start = np.maximum(
                surface.monitor.center[idx] - surface.monitor.size[idx] / 2.0,
                sim_data.simulation.center[idx] - sim_data.simulation.size[idx] / 2.0,
            )
            stop = np.minimum(
                surface.monitor.center[idx] + surface.monitor.size[idx] / 2.0,
                sim_data.simulation.center[idx] + sim_data.simulation.size[idx] / 2.0,
            )
            size = stop - start

            num_pts = int(np.ceil(pts_per_wavelength * size / wavelength))
            points = np.linspace(start, stop, num_pts)
            colocation_points[idx] = points

        currents = currents.colocate(*colocation_points)
        try:
            currents = currents.sel(f=frequency)
        except Exception as e:
            raise SetupError(
                f"Frequency {frequency} not found in fields for monitor '{surface.monitor.name}'."
            ) from e

        return currents

    # pylint:disable=too-many-locals
    def _radiation_vectors_for_surface(
        self, theta: ArrayLikeN2F, phi: ArrayLikeN2F, surface: Near2FarSurface, currents: xr.Dataset
    ):
        """Compute radiation vectors at an angle in spherical coordinates
        for a given set of surface currents and observation angles.

        Parameters
        ----------
        theta : Union[float, List[float], np.ndarray]
            Polar angles (rad) downward from x=y=0 line relative to the local origin.
        phi : Union[float, List[float], np.ndarray]
            Azimuthal (rad) angles from y=z=0 line relative to the local origin.
        surface: :class:`Near2FarSurface`
            :class:`Near2FarSurface` object to use as source of near field.
        currents : xarray.Dataset
            xarray Dataset containing surface currents associated with the surface monitor.

        Returns
        -------
        tuple(numpy.ndarray[float],numpy.ndarray[float],numpy.ndarray[float],numpy.ndarray[float])
            ``N_theta``, ``N_phi``, ``L_theta``, ``L_phi`` radiation vectors for the given surface.
        """

        # make sure that observation points are interpreted w.r.t. the local origin
        pts = [currents[name].values - origin for name, origin in zip(["x", "y", "z"], self.origin)]

        idx_w, idx_uv = surface.monitor.pop_axis((0, 1, 2), axis=surface.axis)
        _, source_names = surface.monitor.pop_axis(("x", "y", "z"), axis=surface.axis)

        idx_u, idx_v = idx_uv
        cmp_1, cmp_2 = source_names

        theta = np.atleast_1d(theta)
        phi = np.atleast_1d(phi)

        sin_theta = np.sin(theta)
        cos_theta = np.cos(theta)
        sin_phi = np.sin(phi)
        cos_phi = np.cos(phi)

        J = np.zeros((3, len(theta), len(phi)), dtype=complex)
        M = np.zeros_like(J)

        def integrate_2d(function, pts_u, pts_v):
            """Trapezoidal integration in two dimensions."""
            return np.trapz(np.trapz(function, pts_u, axis=0), pts_v, axis=0)

        phase = [None] * 3
        propagation_factor = -self.phasor_positive_sign * 1j * self.k

        def integrate_for_one_theta(i_th: int):
            """Perform integration for a given theta angle index"""

            for j_ph in np.arange(len(phi)):

                phase[0] = np.exp(propagation_factor * pts[0] * sin_theta[i_th] * cos_phi[j_ph])
                phase[1] = np.exp(propagation_factor * pts[1] * sin_theta[i_th] * sin_phi[j_ph])
                phase[2] = np.exp(propagation_factor * pts[2] * cos_theta[i_th])

                phase_ij = phase[idx_u][:, None] * phase[idx_v][None, :] * phase[idx_w]

                J[idx_u, i_th, j_ph] = integrate_2d(
                    currents["J" + cmp_1].values * phase_ij, pts[idx_u], pts[idx_v]
                )
                J[idx_v, i_th, j_ph] = integrate_2d(
                    currents["J" + cmp_2].values * phase_ij, pts[idx_u], pts[idx_v]
                )

                M[idx_u, i_th, j_ph] = integrate_2d(
                    currents["M" + cmp_1].values * phase_ij, pts[idx_u], pts[idx_v]
                )
                M[idx_v, i_th, j_ph] = integrate_2d(
                    currents["M" + cmp_2].values * phase_ij, pts[idx_u], pts[idx_v]
                )

        if len(theta) < 2:
            integrate_for_one_theta(0)
        else:
            for i_th in track(
                np.arange(len(theta)),
                description=f"Processing surface monitor '{surface.monitor.name}'...",
            ):
                integrate_for_one_theta(i_th)

        cos_th_cos_phi = cos_theta[:, None] * cos_phi[None, :]
        cos_th_sin_phi = cos_theta[:, None] * sin_phi[None, :]

        # N_theta (8.33a)
        N_theta = J[0] * cos_th_cos_phi + J[1] * cos_th_sin_phi - J[2] * sin_theta[:, None]

        # N_phi (8.33b)
        N_phi = -J[0] * sin_phi[None, :] + J[1] * cos_phi[None, :]

        # L_theta  (8.34a)
        L_theta = M[0] * cos_th_cos_phi + M[1] * cos_th_sin_phi - M[2] * sin_theta[:, None]

        # L_phi  (8.34b)
        L_phi = -M[0] * sin_phi[None, :] + M[1] * cos_phi[None, :]

        return N_theta, N_phi, L_theta, L_phi

    def _radiation_vectors(self, theta: ArrayLikeN2F, phi: ArrayLikeN2F):
        """Compute radiation vectors at an angle in spherical coordinates.

        Parameters
        ----------
        theta : Union[float, List[float], np.ndarray]
            Polar angles (rad) downward from x=y=0 line relative to the local origin.
        phi : Union[float, List[float], np.ndarray]
            Azimuthal (rad) angles from y=z=0 line relative to the local origin.

        Returns
        -------
        tuple(numpy.ndarray[float],numpy.ndarray[float],numpy.ndarray[float],numpy.ndarray[float])
            ``N_theta``, ``N_phi``, ``L_theta``, ``L_phi`` radiation vectors.
        """

        # compute radiation vectors for the dataset associated with each monitor
        N_theta = np.zeros((len(theta), len(phi)), dtype=complex)
        N_phi = np.zeros_like(N_theta)
        L_theta = np.zeros_like(N_theta)
        L_phi = np.zeros_like(N_theta)

        for surface in self.surfaces:
            _N_th, _N_ph, _L_th, _L_ph = self._radiation_vectors_for_surface(
                theta, phi, surface, self.currents[surface.monitor.name]
            )
            N_theta += _N_th
            N_phi += _N_ph
            L_theta += _L_th
            L_phi += _L_ph

        return N_theta, N_phi, L_theta, L_phi

    def fields_spherical(self, r: float, theta: ArrayLikeN2F, phi: ArrayLikeN2F) -> xr.Dataset:
        """Get fields at a point relative to monitor center in spherical coordinates.

        Parameters
        ----------
        r : float
            (micron) radial distance relative to monitor center.
        theta : Union[float, List[float], np.ndarray]
            (radian) polar angles downward from x=y=0 relative to the local origin.
        phi : Union[float, List[float], np.ndarray]
            (radian) azimuthal angles from y=z=0 line relative to the local origin.

        Returns
        -------
        xarray.Dataset
            xarray dataset containing (Er, Etheta, Ephi), (Hr, Htheta, Hphi)
            in polar coordinates.
        """

        theta = np.atleast_1d(theta)
        phi = np.atleast_1d(phi)

        # project radiation vectors to distance r away for given angles
        N_theta, N_phi, L_theta, L_phi = self._radiation_vectors(theta, phi)

        k = self.k
        eta = self.eta

        scalar_proj_r = (
            -self.phasor_positive_sign
            * 1j
            * k
            * np.exp(self.phasor_positive_sign * 1j * k * r)
            / (4 * np.pi * r)
        )

        # assemble E felds
        Et_array = -scalar_proj_r * (L_phi + eta * N_theta)
        Ep_array = scalar_proj_r * (L_theta - eta * N_phi)
        Er_array = np.zeros_like(Ep_array)

        # assemble H fields
        Ht_array = -Ep_array / eta
        Hp_array = Et_array / eta
        Hr_array = np.zeros_like(Hp_array)

        dims = ("r", "theta", "phi")
        coords = {"r": [r], "theta": theta, "phi": phi}

        Er = xr.DataArray(data=Er_array[None, ...], coords=coords, dims=dims)
        Et = xr.DataArray(data=Et_array[None, ...], coords=coords, dims=dims)
        Ep = xr.DataArray(data=Ep_array[None, ...], coords=coords, dims=dims)

        Hr = xr.DataArray(data=Hr_array[None, ...], coords=coords, dims=dims)
        Ht = xr.DataArray(data=Ht_array[None, ...], coords=coords, dims=dims)
        Hp = xr.DataArray(data=Hp_array[None, ...], coords=coords, dims=dims)

        field_data = xr.Dataset(
            {"E_r": Er, "E_theta": Et, "E_phi": Ep, "H_r": Hr, "H_theta": Ht, "H_phi": Hp}
        )

        return field_data

    def fields_cartesian(self, x: ArrayLikeN2F, y: ArrayLikeN2F, z: ArrayLikeN2F) -> xr.Dataset:
        """Get fields at a point relative to monitor center in cartesian coordinates.

        Parameters
        ----------
        x : Union[float, List[float], np.ndarray]
            (micron) x positions relative to the local origin.
        y : Union[float, List[float], np.ndarray]
            (micron) y positions relative to the local origin.
        z : Union[float, List[float], np.ndarray]
            (micron) z positions relative to the local origin.

        Returns
        -------
        xarray.Dataset
            xarray dataset containing (Ex, Ey, Ez), (Hx, Hy, Hz) in cartesian coordinates.
        """

        x, y, z = [np.atleast_1d(x), np.atleast_1d(y), np.atleast_1d(z)]

        Ex_data = np.zeros((len(x), len(y), len(z)), dtype=complex)
        Ey_data = np.zeros_like(Ex_data)
        Ez_data = np.zeros_like(Ex_data)

        Hx_data = np.zeros_like(Ex_data)
        Hy_data = np.zeros_like(Ex_data)
        Hz_data = np.zeros_like(Ex_data)

        for i in track(np.arange(len(x)), description="Computing far fields"):
            _x = x[i]
            for j in np.arange(len(y)):
                _y = y[j]
                for k in np.arange(len(z)):
                    _z = z[k]

                    r, theta, phi = self._car_2_sph(_x, _y, _z)
                    _field_data = self.fields_spherical(r, theta, phi)

                    Er, Et, Ep = [_field_data[comp].values for comp in ["E_r", "E_theta", "E_phi"]]
                    Hr, Ht, Hp = [_field_data[comp].values for comp in ["H_r", "H_theta", "H_phi"]]

                    Ex_data[i, j, k], Ey_data[i, j, k], Ez_data[i, j, k] = self._sph_2_car_field(
                        Er, Et, Ep, theta, phi
                    )
                    Hx_data[i, j, k], Hy_data[i, j, k], Hz_data[i, j, k] = self._sph_2_car_field(
                        Hr, Ht, Hp, theta, phi
                    )

        dims = ("x", "y", "z")
        coords = {"x": x, "y": y, "z": z}

        Ex = xr.DataArray(data=Ex_data, coords=coords, dims=dims)
        Ey = xr.DataArray(data=Ey_data, coords=coords, dims=dims)
        Ez = xr.DataArray(data=Ez_data, coords=coords, dims=dims)

        Hx = xr.DataArray(data=Hx_data, coords=coords, dims=dims)
        Hy = xr.DataArray(data=Hy_data, coords=coords, dims=dims)
        Hz = xr.DataArray(data=Hz_data, coords=coords, dims=dims)

        field_data = xr.Dataset({"Ex": Ex, "Ey": Ey, "Ez": Ez, "Hx": Hx, "Hy": Hy, "Hz": Hz})

        return field_data

    def power_spherical(self, r: float, theta: ArrayLikeN2F, phi: ArrayLikeN2F) -> xr.DataArray:
        """Get power scattered to a point relative to the local origin in spherical coordinates.

        Parameters
        ----------
        r : float
            (micron) radial distance relative to the local origin.
        theta : Union[float, List[float], np.ndarray]
            (radian) polar angles downward from x=y=0 relative to the local origin.
        phi : Union[float, List[float], np.ndarray]
            (radian) azimuthal angles from y=z=0 line relative to the local origin.

        Returns
        -------
        power : xarray.DataArray
            Power at points relative to the local origin.
        """

        theta = np.atleast_1d(theta)
        phi = np.atleast_1d(phi)

        field_data = self.fields_spherical(r, theta, phi)
        Et, Ep = [field_data[comp].values for comp in ["E_theta", "E_phi"]]
        Ht, Hp = [field_data[comp].values for comp in ["H_theta", "H_phi"]]
        power_theta = 0.5 * np.real(Et * np.conj(Hp))
        power_phi = 0.5 * np.real(-Ep * np.conj(Ht))
        power_data = power_theta + power_phi

        dims = ("r", "theta", "phi")
        coords = {"r": [r], "theta": theta, "phi": phi}

        return xr.DataArray(data=power_data, coords=coords, dims=dims)

    def power_cartesian(self, x: ArrayLikeN2F, y: ArrayLikeN2F, z: ArrayLikeN2F) -> xr.DataArray:
        """Get power scattered to a point relative to the local origin in cartesian coordinates.

        Parameters
        ----------
        x : Union[float, List[float], np.ndarray]
            (micron) x distances relative to the local origin.
        y : Union[float, List[float], np.ndarray]
            (micron) y distances relative to the local origin.
        z : Union[float, List[float], np.ndarray]
            (micron) z distances relative to the local origin.

        Returns
        -------
        power : xarray.DataArray
            Power at points relative to the local origin.
        """

        x, y, z = [np.atleast_1d(x), np.atleast_1d(y), np.atleast_1d(z)]

        power_data = np.zeros((len(x), len(y), len(z)))

        for i in track(np.arange(len(x)), description="Computing far field power"):
            _x = x[i]
            for j in np.arange(len(y)):
                _y = y[j]
                for k in np.arange(len(z)):
                    _z = z[k]

                    r, theta, phi = self._car_2_sph(_x, _y, _z)
                    power_data[i, j, k] = self.power_spherical(r, theta, phi).values

        dims = ("x", "y", "z")
        coords = {"x": x, "y": y, "z": z}

        return xr.DataArray(data=power_data, coords=coords, dims=dims)

    def radar_cross_section(self, theta: ArrayLikeN2F, phi: ArrayLikeN2F) -> xr.DataArray:
        """Get radar cross section at a point relative to the local origin in
        units of incident power.

        Parameters
        ----------
        theta : Union[float, List[float], np.ndarray]
            (radian) polar angles downward from x=y=0 relative to the local origin.
        phi : Union[float, List[float], np.ndarray]
            (radian) azimuthal angles from y=z=0 line relative to the local origin.

        Returns
        -------
        RCS : xarray.DataArray
            Radar cross section at angles relative to the local origin.
        """

        theta = np.atleast_1d(theta)
        phi = np.atleast_1d(phi)

        _, index_k = self.nk
        if index_k != 0.0:
            raise SetupError("Can't compute RCS for a lossy background medium.")

        # set observation angles relative to the local origin
        N_theta, N_phi, L_theta, L_phi = self._radiation_vectors(theta, phi)

        # wave number and wave impedance must be real since index_k is forced to be 0
        eta = np.real(self.eta)
        k = np.real(self.k)

        constant = k**2 / (8 * np.pi * eta)
        term1 = np.abs(L_phi + eta * N_theta) ** 2
        term2 = np.abs(L_theta - eta * N_phi) ** 2
        rcs_data = constant * (term1 + term2)

        dims = ("theta", "phi")
        coords = {"theta": theta, "phi": phi}

        return xr.DataArray(data=rcs_data, coords=coords, dims=dims)

    @staticmethod
    def _car_2_sph(x, y, z):
        """
        Parameters
        ----------
        x : float
            x coordinate.
        y : float
            y coordinate.
        z : float
            z coordinate.

        Returns
        -------
        tuple
            r, theta, and phi in spherical coordinates.
        """
        r = np.sqrt(x**2 + y**2 + z**2)
        theta = np.arccos(z / r)
        phi = np.arctan2(y, x)
        return r, theta, phi

    @staticmethod
    def _sph_2_car(r, theta, phi):
        """coordinates only

        Parameters
        ----------
        r : float
            radius.
        theta : float
            polar angle (rad) downward from x=y=0 line.
        phi : float
            azimuthal (rad) angle from y=z=0 line.

        Returns
        -------
        tuple
            x, y, and z in cartesian coordinates.
        """
        r_sin_theta = r * np.sin(theta)
        x = r_sin_theta * np.cos(phi)
        y = r_sin_theta * np.sin(phi)
        z = r * np.cos(theta)
        return x, y, z

    @staticmethod
    def _sph_2_car_field(Ar, Atheta, Aphi, theta, phi):
        """Convert vector field components in spherical coordinates to cartesian.

        Parameters
        ----------
        Ar : float
            radial component of vector A.
        Atheta : float
            polar angle component of vector A.
        Aphi : float
            azimuthal angle component of vector A.
        theta : float
            polar angle (rad) of location of A.
        phi : float
            azimuthal angle (rad) of location of A.

        Returns
        -------
        tuple
            x, y, and z components of A in cartesian coordinates.
        """
        sin_theta = np.sin(theta)
        cos_theta = np.cos(theta)
        sin_phi = np.sin(phi)
        cos_phi = np.cos(phi)
        Ax = Ar * sin_theta * cos_phi + Atheta * cos_theta * cos_phi - Aphi * sin_phi
        Ay = Ar * sin_theta * sin_phi + Atheta * cos_theta * sin_phi + Aphi * cos_phi
        Az = Ar * cos_theta - Atheta * sin_theta
        return Ax, Ay, Az



import pytest
import numpy as np
import pydantic

import gdspy
import tidy3d as td

from tidy3d.plugins import DispersionFitter
from tidy3d.plugins.webplots import SimulationPlotly, SimulationDataApp
from tidy3d.plugins import ModeSolver
from tidy3d.plugins import Near2Far
from tidy3d import FieldData, ScalarFieldData, FieldMonitor
from tidy3d.plugins.smatrix.smatrix import Port
from tidy3d.plugins.smatrix.smatrix import ComponentModeler
from .utils import clear_tmp


def test_near2far():
    """make sure Near2Far runs"""

    center = (0, 0, 0)
    size = (2, 2, 2)
    f0 = 1
    monitors = FieldMonitor(size=size, center=center, freqs=[f0], name="near_field").surfaces()

    sim_size = (5, 5, 5)
    sim = td.Simulation(
        size=sim_size,
        grid_spec=td.GridSpec.auto(wavelength=td.C_0 / f0),
        monitors=monitors,
        run_time=1e-12,
    )

    def rand_data():
        return ScalarFieldData(
            x=np.linspace(-1, 1, 10),
            y=np.linspace(-1, 1, 10),
            z=np.linspace(-1, 1, 10),
            f=[f0],
            values=np.random.random((10, 10, 10, 1)),
        )

    fields = ["Ex", "Ey", "Ez", "Hx", "Hy", "Hz"]
    data_dict = {field: rand_data() for field in fields}
    field_data = FieldData(data_dict=data_dict)

    data_dict_mon = {mon.name: field_data for mon in monitors}
    sim_data = td.SimulationData(simulation=sim, monitor_data=data_dict_mon)

    n2f = Near2Far.from_surface_monitors(
        sim_data=sim_data,
        monitors=monitors,
        normal_dirs=["-", "+", "-", "+", "-", "+"],
        frequency=f0,
    )

    # single inputs
    n2f.radar_cross_section(1, 1)
    n2f.power_spherical(1, 1, 1)
    n2f.power_cartesian(1, 1, 1)
    n2f.fields_spherical(1, 1, 1)
    n2f.fields_cartesian(1, 1, 1)

    # vectorized inputs
    pts1 = [0, 1]
    pts2 = [0, 1, 2]
    pts3 = [3, 4, 5]
    n2f.radar_cross_section(pts1, pts2)
    n2f.power_spherical(1, pts2, pts3)
    n2f.power_cartesian(pts1, pts2, pts3)
    n2f.fields_spherical(1, pts2, pts3)
    n2f.fields_cartesian(pts1, pts2, pts3)


def test_mode_solver():
    """make sure mode solver runs"""
    waveguide = td.Structure(
        geometry=td.Box(size=(100, 0.5, 0.5)), medium=td.Medium(permittivity=4.0)
    )
    simulation = td.Simulation(
        size=(2, 2, 2),
        grid_spec=td.GridSpec(wavelength=1.0),
        structures=[waveguide],
        run_time=1e-12,
    )
    plane = td.Box(center=(0, 0, 0), size=(0, 1, 1))
    mode_spec = td.ModeSpec(
        num_modes=3,
        target_neff=2.0,
        bend_radius=3.0,
        bend_axis=0,
        num_pml=(10, 10),
    )
    ms = ModeSolver(
        simulation=simulation, plane=plane, mode_spec=mode_spec, freqs=[td.constants.C_0 / 1.0]
    )
    modes = ms.solve()


def _test_coeffs():
    """make sure pack_coeffs and unpack_coeffs are reciprocal"""
    num_poles = 10
    coeffs = np.random.random(4 * num_poles)
    a, c = _unpack_coeffs(coeffs)
    coeffs_ = _pack_coeffs(a, c)
    a_, c_ = _unpack_coeffs(coeffs_)
    assert np.allclose(coeffs, coeffs_)
    assert np.allclose(a, a_)
    assert np.allclose(c, c_)


def _test_pole_coeffs():
    """make sure coeffs_to_poles and poles_to_coeffs are reciprocal"""
    num_poles = 10
    coeffs = np.random.random(4 * num_poles)
    poles = _coeffs_to_poles(coeffs)
    coeffs_ = _poles_to_coeffs(poles)
    poles_ = _coeffs_to_poles(coeffs_)
    assert np.allclose(coeffs, coeffs_)


@clear_tmp
def test_dispersion():
    """performs a fit on some random data"""
    num_data = 10
    n_data = np.random.random(num_data)
    wvls = np.linspace(1, 2, num_data)
    fitter = DispersionFitter(wvl_um=wvls, n_data=n_data)
    medium, rms = fitter._fit_single()
    medium, rms = fitter.fit(num_tries=2)
    medium.to_file("tests/tmp/medium_fit.json")

    k_data = np.random.random(num_data)
    fitter = DispersionFitter(wvl_um=wvls, n_data=n_data, k_data=k_data)


def test_dispersion_load():
    """loads dispersion model from nk data file"""
    fitter = DispersionFitter.from_file("tests/data/nk_data.csv", skiprows=1, delimiter=",")
    medium, rms = fitter.fit(num_tries=20)


def test_dispersion_plot():
    """plots a medium fit from file"""
    fitter = DispersionFitter.from_file("tests/data/nk_data.csv", skiprows=1, delimiter=",")
    medium, rms = fitter.fit(num_tries=20)
    fitter.plot(medium)


def test_dispersion_set_wvg_range():
    """set wavelength range function"""
    num_data = 50
    n_data = np.random.random(num_data)
    wvls = np.linspace(1, 2, num_data)
    fitter = DispersionFitter(wvl_um=wvls, n_data=n_data)

    wvl_min = np.random.random(1)[0] * 0.5 + 1
    wvl_max = wvl_min + 0.5
    fitter.wvl_range = [wvl_min, wvl_max]
    assert len(fitter.freqs) < num_data
    medium, rms = fitter.fit(num_tries=2)


def test_plotly():
    """Tests plotly plotting."""
    s = td.Simulation(size=(1, 1, 1), grid_spec=td.GridSpec.auto(wavelength=1), run_time=1e-12)
    sp = SimulationPlotly(simulation=s)
    fig = sp.plotly(x=0)


def test_app():
    """Tests plotly app."""
    center = (0, 0, 0)
    size = (2, 2, 2)
    f0 = 1
    monitors = FieldMonitor(size=size, center=center, freqs=[f0], name="near_field").surfaces()

    sim_size = (5, 5, 5)
    sim = td.Simulation(
        size=sim_size,
        grid_spec=td.GridSpec.auto(wavelength=td.C_0 / f0),
        monitors=monitors,
        run_time=1e-12,
    )

    def rand_data():
        return ScalarFieldData(
            x=np.linspace(-1, 1, 10),
            y=np.linspace(-1, 1, 10),
            z=np.linspace(-1, 1, 10),
            f=[f0],
            values=np.random.random((10, 10, 10, 1)),
        )

    fields = ["Ex", "Ey", "Ez", "Hx", "Hy", "Hz"]
    data_dict = {field: rand_data() for field in fields}
    field_data = FieldData(data_dict=data_dict)

    data_dict_mon = {mon.name: field_data for mon in monitors}
    sim_data = td.SimulationData(simulation=sim, monitor_data=data_dict_mon)

    app = SimulationDataApp(sim_data=sim_data)
    _app = app.make_app()


def test_component_modeler():
    """Tests S matrix loading."""
    # wavelength / frequency
    lambda0 = 1.550  # all length scales in microns
    freq0 = td.constants.C_0 / lambda0
    fwidth = freq0 / 10

    # Spatial grid specification
    grid_spec = td.GridSpec.auto(min_steps_per_wvl=14, wavelength=lambda0)

    # Permittivity of waveguide and substrate
    wg_n = 3.48
    sub_n = 1.45
    mat_wg = td.Medium(permittivity=wg_n**2)
    mat_sub = td.Medium(permittivity=sub_n**2)

    # Waveguide dimensions

    # Waveguide height
    wg_height = 0.22
    # Waveguide width
    wg_width = 1.0
    # Waveguide separation in the beginning/end
    wg_spacing_in = 8
    # length of coupling region (um)
    coup_length = 6.0
    # spacing between waveguides in coupling region (um)
    wg_spacing_coup = 0.05
    # Total device length along propagation direction
    device_length = 100
    # Length of the bend region
    bend_length = 16
    # Straight waveguide sections on each side
    straight_wg_length = 4
    # space between waveguide and PML
    pml_spacing = 2

    def bend_pts(bend_length, width, npts=10):
        """Set of points describing a tanh bend from (0, 0) to (length, width)"""
        x = np.linspace(0, bend_length, npts)
        y = width * (1 + np.tanh(6 * (x / bend_length - 0.5))) / 2
        return np.stack((x, y), axis=1)

    def arm_pts(length, width, coup_length, bend_length, npts_bend=30):
        """Set of points defining one arm of an integrated coupler"""
        ### Make the right half of the coupler arm first
        # Make bend and offset by coup_length/2
        bend = bend_pts(bend_length, width, npts_bend)
        bend[:, 0] += coup_length / 2
        # Add starting point as (0, 0)
        right_half = np.concatenate(([[0, 0]], bend))
        # Add an extra point to make sure waveguide is straight past the bend
        right_half = np.concatenate((right_half, [[right_half[-1, 0] + 0.1, width]]))
        # Add end point as (length/2, width)
        right_half = np.concatenate((right_half, [[length / 2, width]]))

        # Make the left half by reflecting and omitting the (0, 0) point
        left_half = np.copy(right_half)[1:, :]
        left_half[:, 0] = -left_half[::-1, 0]
        left_half[:, 1] = left_half[::-1, 1]

        return np.concatenate((left_half, right_half), axis=0)

    def make_coupler(
        length, wg_spacing_in, wg_width, wg_spacing_coup, coup_length, bend_length, npts_bend=30
    ):
        """Make an integrated coupler using the gdspy FlexPath object."""

        # Compute one arm of the coupler
        arm_width = (wg_spacing_in - wg_width - wg_spacing_coup) / 2
        arm = arm_pts(length, arm_width, coup_length, bend_length, npts_bend)
        # Reflect and offset bottom arm
        coup_bot = np.copy(arm)
        coup_bot[:, 1] = -coup_bot[::-1, 1] - wg_width / 2 - wg_spacing_coup / 2
        # Offset top arm
        coup_top = np.copy(arm)
        coup_top[:, 1] += wg_width / 2 + wg_spacing_coup / 2

        # Create waveguides as GDS paths
        path_bot = gdspy.FlexPath(coup_bot, wg_width, layer=1, datatype=0)
        path_top = gdspy.FlexPath(coup_top, wg_width, layer=1, datatype=1)

        return [path_bot, path_top]

    gdspy.current_library = gdspy.GdsLibrary()
    lib = gdspy.GdsLibrary()

    # Geometry must be placed in GDS cells to import into Tidy3D
    coup_cell = lib.new_cell("Coupler")

    substrate = gdspy.Rectangle(
        (-device_length / 2, -wg_spacing_in / 2 - 10),
        (device_length / 2, wg_spacing_in / 2 + 10),
        layer=0,
    )
    coup_cell.add(substrate)

    # Add the coupler to a gdspy cell
    gds_coup = make_coupler(
        device_length, wg_spacing_in, wg_width, wg_spacing_coup, coup_length, bend_length
    )
    coup_cell.add(gds_coup)

    # Substrate
    [oxide_geo] = td.PolySlab.from_gds(
        gds_cell=coup_cell, gds_layer=0, gds_dtype=0, slab_bounds=(-10, 0), axis=2
    )

    oxide = td.Structure(geometry=oxide_geo, medium=mat_sub)

    # Waveguides (import all datatypes if gds_dtype not specified)
    coupler1_geo, coupler2_geo = td.PolySlab.from_gds(
        gds_cell=coup_cell, gds_layer=1, slab_bounds=(0, wg_height), axis=2
    )

    coupler1 = td.Structure(geometry=coupler1_geo, medium=mat_wg)

    coupler2 = td.Structure(geometry=coupler2_geo, medium=mat_wg)

    # Simulation size along propagation direction
    sim_length = 2 * straight_wg_length + 2 * bend_length + coup_length

    # Spacing between waveguides and PML
    sim_size = [sim_length, wg_spacing_in + wg_width + 2 * pml_spacing, wg_height + 2 * pml_spacing]

    # source
    src_pos = sim_length / 2 - straight_wg_length / 2

    # in-plane field monitor (optional, increases required data storage)
    domain_monitor = td.FieldMonitor(
        center=[0, 0, wg_height / 2], size=[td.inf, td.inf, 0], freqs=[freq0], name="field"
    )

    # initialize the simulation
    sim = td.Simulation(
        size=sim_size,
        grid_spec=grid_spec,
        structures=[oxide, coupler1, coupler2],
        sources=[],
        monitors=[domain_monitor],
        run_time=20 / fwidth,
        pml_layers=3 * [td.PML()],
    )

    port_right_top = Port(
        center=[src_pos, wg_spacing_in / 2, wg_height / 2],
        size=[0, 4, 2],
        mode_spec=td.ModeSpec(num_modes=2),
        direction="-",
        name="right_top",
    )

    port_right_bot = Port(
        center=[src_pos, -wg_spacing_in / 2, wg_height / 2],
        size=[0, 4, 2],
        mode_spec=td.ModeSpec(num_modes=2),
        direction="-",
        name="right_bot",
    )

    port_left_top = Port(
        center=[-src_pos, wg_spacing_in / 2, wg_height / 2],
        size=[0, 4, 2],
        mode_spec=td.ModeSpec(num_modes=2),
        direction="+",
        name="left_top",
    )

    port_left_bot = Port(
        center=[-src_pos, -wg_spacing_in / 2, wg_height / 2],
        size=[0, 4, 2],
        mode_spec=td.ModeSpec(num_modes=2),
        direction="+",
        name="left_bot",
    )

    ports = [port_right_top, port_right_bot, port_left_top, port_left_bot]
    modeler = ComponentModeler(simulation=sim, ports=ports, freq=freq0)



import numpy as np
from tidy3d.components.data import *
import tidy3d as td
from .utils import clear_tmp


def test_scalar_field_data():
    f = np.linspace(1e14, 2e14, 1001)
    x = np.linspace(-1, 1, 10)
    y = np.linspace(-2, 2, 20)
    z = np.linspace(0, 0, 1)
    values = (1 + 1j) * np.random.random((len(x), len(y), len(z), len(f)))
    data = ScalarFieldData(values=values, x=x, y=y, z=z, f=f)
    data.data


def test_scalar_field_time_data():
    t = np.linspace(0, 1e-12, 1001)
    x = np.linspace(-1, 1, 10)
    y = np.linspace(-2, 2, 20)
    z = np.linspace(0, 0, 1)
    values = np.random.random((len(x), len(y), len(z), len(t)))
    data = ScalarFieldTimeData(values=values, x=x, y=y, z=z, t=t)
    data.data


def test_scalar_permittivity_data():
    f = np.linspace(1e14, 2e14, 1001)
    x = np.linspace(-1, 1, 10)
    y = np.linspace(-2, 2, 20)
    z = np.linspace(0, 0, 1)
    values = (1 + 1j) * np.random.random((len(x), len(y), len(z), len(f)))
    data = ScalarPermittivityData(values=values, x=x, y=y, z=z, f=f)
    data.data


def test_mode_amps_data():
    f = np.linspace(2e14, 3e14, 1001)
    mode_index = np.arange(1, 3)
    values = (1 + 1j) * np.random.random((2, len(f), len(mode_index)))
    data = ModeAmpsData(values=values, direction=["+", "-"], mode_index=mode_index, f=f)
    data.data


def test_mode_index_data():
    f = np.linspace(2e14, 3e14, 1001)
    mode_index = np.arange(1, 3)
    values = (1 + 1j) * np.random.random((len(f), len(mode_index)))
    data = ModeIndexData(values=values, f=f, mode_index=np.arange(1, 3))
    data.n_eff
    data.k_eff
    data.n_complex
    data.data


def test_field_data():
    f = np.linspace(1e14, 2e14, 1001)
    x = np.linspace(-1, 1, 10)
    y = np.linspace(-2, 2, 20)
    z = np.linspace(0, 0, 1)
    values = (1 + 1j) * np.random.random((len(x), len(y), len(z), len(f)))
    field = ScalarFieldData(values=values, x=x, y=y, z=z, f=f)
    data = FieldData(data_dict={"Ex": field, "Ey": field})
    data.data


def test_field_time_data():
    t = np.linspace(0, 1e-12, 1001)
    x = np.linspace(-1, 1, 10)
    y = np.linspace(-2, 2, 20)
    z = np.linspace(0, 0, 1)
    values = np.random.random((len(x), len(y), len(z), len(t)))
    field = ScalarFieldTimeData(values=values, x=x, y=y, z=z, t=t)
    data = FieldTimeData(data_dict={"Ex": field, "Ey": field})
    data.data


def test_permittivity_data():
    f = np.linspace(1e14, 2e14, 1001)
    x = np.linspace(-1, 1, 10)
    y = np.linspace(-2, 2, 20)
    z = np.linspace(0, 0, 1)
    values = (1 + 1j) * np.random.random((len(x), len(y), len(z), len(f)))
    eps = ScalarPermittivityData(values=values, x=x, y=y, z=z, f=f)
    data = PermittivityData(data_dict={"eps_xx": eps, "eps_yy": eps, "eps_zz": eps})
    data.eps_xx
    data.eps_yy
    data.eps_zz
    data.data


def test_flux_data():
    f = np.linspace(2e14, 3e14, 1001)
    values = np.random.random((len(f),))
    data = FluxData(values=values, f=f)
    data.data


def test_mode_field_data():
    f = np.linspace(1e14, 2e14, 1001)
    x = np.linspace(-1, 1, 10)
    y = np.linspace(-2, 2, 20)
    z = np.linspace(0, 0, 1)
    mode_index = np.arange(0, 4)
    values = (1 + 1j) * np.random.random((len(x), len(y), len(z), len(f), len(mode_index)))
    field = ScalarModeFieldData(values=values, x=x, y=y, z=z, f=f, mode_index=mode_index)
    data = ModeFieldData(data_dict={"Ex": field, "Ey": field})
    data.sel_mode_index(1)


@clear_tmp
def test_sim_data():

    center = (0, 0, 0)
    size = (2, 2, 2)
    f0 = 1
    monitor = td.FieldMonitor(size=size, center=center, freqs=[f0], name="test")

    sim_size = (5, 5, 5)
    sim = td.Simulation(
        size=sim_size,
        grid_spec=td.GridSpec.auto(wavelength=td.C_0 / f0),
        monitors=[monitor],
        run_time=1e-12,
        sources=[
            td.PointDipole(
                center=(0, 0, 0),
                source_time=td.GaussianPulse(freq0=2e14, fwidth=2e13),
                polarization="Ex",
            )
        ],
    )

    def rand_data():
        return ScalarFieldData(
            x=np.linspace(-1, 1, 10),
            y=np.linspace(-1, 1, 10),
            z=np.linspace(-1, 1, 10),
            f=[f0],
            values=np.random.random((10, 10, 10, 1)),
        )

    fields = ["Ex", "Ey", "Ez", "Hx", "Hy", "Hz"]
    data_dict = {field: rand_data() for field in fields}
    field_data = FieldData(data_dict=data_dict)

    sim_data = SimulationData(
        simulation=sim, monitor_data={"test": field_data}, log_string="field_decay :1e-4"
    )

    sim_data.plot_field("test", "Ex", val="real", x=0, freq=f0)
    sim_data.at_centers("test")
    sim_data.normalized
    sim_data.diverged
    sim_data.final_decay_value
    sim_data.log
    sim_data["test"].Ex
    sim_data["test"].Ey
    sim_data["test"].Ez
    sim_data["test"].Ex
    sim_data["test"].Ey
    sim_data["test"].Ez
    sim_data.normalize()
    sim_data.to_file("tests/tmp/sim_data.hdf5")
    sim_data2 = SimulationData.from_file("tests/tmp/sim_data.hdf5")
    # assert sim_data == sim_data2


def test_symmetries():
    f = np.linspace(1e14, 2e14, 1001)
    x = np.linspace(-1, 1, 10)
    y = np.linspace(-2, 2, 20)
    z = np.linspace(0, 0, 1)
    values = (1 + 1j) * np.random.random((len(x), len(y), len(z), len(f)))
    field = ScalarFieldData(values=values, x=x, y=y, z=z, f=f)
    data = FieldData(
        data_dict={"Ex": field, "Ey": field}, symmetry=(1, -1, 0), symmetry_center=(0, 0, 0)
    )
    data.expand_syms



from typing import Dict
import pytest
import numpy as np
import pydantic
from shapely.geometry import Polygon, Point

import tidy3d as td
from tidy3d.log import ValidationError, SetupError

np.random.seed(4)


def setup_polyslab(vertices, dilation, angle, bounds):
    s = td.PolySlab(
        vertices=vertices,
        slab_bounds=bounds,
        axis=2,
        dilation=dilation,
        sidewall_angle=angle,
    )
    return s


def minimal_edge_length(vertices):
    """compute the minimal edge length in a polygon"""
    vs = vertices.T.copy()
    vsp = np.roll(vs.copy(), axis=-1, shift=-1)
    edge_length = np.linalg.norm(vsp - vs, axis=0)
    return np.min(edge_length)


def convert_valid_polygon(vertices):
    """Given vertices that might have intersecting eduges, generate
    vertices of a valid polygon
    """
    poly = Polygon(vertices).buffer(0)  # make sure no intersecting edges
    if type(poly) is not Polygon:
        poly = poly[0]

    vertices_n = np.array(poly.exterior.coords[:])
    return vertices_n


def validate_poly_bound(poly):
    """validate bound based polyslab's base and top polygon"""
    xmin1, ymin1 = np.amin(poly.base_polygon, axis=0)
    xmax1, ymax1 = np.amax(poly.base_polygon, axis=0)

    xmin2, ymin2 = np.amin(poly.top_polygon, axis=0)
    xmax2, ymax2 = np.amax(poly.top_polygon, axis=0)

    xmin, ymin = min(xmin1, xmin2), min(ymin1, ymin2)
    xmax, ymax = max(xmax1, xmax2), max(ymax1, ymax2)

    bound_tidy = poly.bounds
    assert bound_tidy[0][0] <= xmin
    assert bound_tidy[0][1] <= ymin
    assert bound_tidy[1][0] >= xmax
    assert bound_tidy[1][1] >= ymax


# default values
bounds = (0, 0.5)
dilation = 0.0
angle = 0


def test_remove_duplicate():
    """
    Make sure redundant neighboring vertices are removed
    """
    vertices = np.random.random((10, 2))
    vertices[0, :] = vertices[9, :]
    vertices[1, :] = vertices[0, :]
    vertices[5, :] = vertices[6, :]

    vertices = td.PolySlab._remove_duplicate_vertices(vertices)
    assert vertices.shape[0] == 7


def test_valid_polygon():
    """No intersecting edges"""

    # area = 0
    vertices = ((0, 0), (1, 0), (2, 0))
    with pytest.raises(SetupError) as e_info:
        s = setup_polyslab(vertices, dilation, angle, bounds)

    # only two points
    vertices = ((0, 0), (1, 0), (1, 0))
    with pytest.raises(SetupError) as e_info:
        s = setup_polyslab(vertices, dilation, angle, bounds)

    # intersecting edges
    vertices = ((0, 0), (1, 0), (1, 1), (0, 1), (0.5, -1))

    with pytest.raises(SetupError) as e_info:
        s = setup_polyslab(vertices, dilation, angle, bounds)


def test_crossing_square():
    """
    Vertices crossing detection for a simple square
    """
    vertices = ((0, 0), (1, 0), (1, -1), (0, -1))
    dilation = 0.0
    angle = np.pi / 4
    s = setup_polyslab(vertices, dilation, angle, bounds)

    # dilation too significant
    dilation = -1.1
    angle = 0
    with pytest.raises(SetupError) as e_info:
        s = setup_polyslab(vertices, dilation, angle, bounds)

    # angle too large
    dilation = 0
    angle = np.pi / 3
    with pytest.raises(SetupError) as e_info:
        s = setup_polyslab(vertices, dilation, angle, bounds)

    # combines both
    dilation = -0.1
    angle = np.pi / 4
    with pytest.raises(SetupError) as e_info:
        s = setup_polyslab(vertices, dilation, angle, bounds)


def test_max_erosion_polygon():
    """
    Maximal erosion distance validation
    """
    N = 10  # number of vertices
    for i in range(50):
        vertices = convert_valid_polygon(np.random.random((N, 2)) * 10)

        dilation = 0
        angle = 0
        bounds = (0, 0.5)
        s = setup_polyslab(vertices, dilation, angle, bounds)

        # compute maximal allowed erosion distance
        _, max_dist = s._crossing_detection(s.base_polygon, -100)

        # verify it is indeed maximal allowed
        dilation = -max_dist + 1e-10
        # avoid vertex-edge crossing case
        try:
            s = setup_polyslab(vertices, dilation, angle, bounds)
        except:
            continue
        assert np.isclose(minimal_edge_length(s.base_polygon), 0, atol=1e-4)

        # verify it is indeed maximal allowed
        dilation = 0.0
        bounds = (0, max_dist - 1e-10)
        angle = np.pi / 4

        # avoid vertex-edge crossing case
        try:
            s = setup_polyslab(vertices, dilation, angle, bounds)
        except:
            continue
        assert np.isclose(minimal_edge_length(s.top_polygon), 0, atol=1e-4)


def test_shift_height():
    """Make sure a list of height where the plane will intersect with the vertices
    works properly
    """
    N = 10  # number of vertices
    Lx = 10.0
    for i in range(50):
        vertices = convert_valid_polygon(np.random.random((N, 2)) * Lx)
        dilation = 0
        angle = 0
        bounds = (0, 1)
        s = setup_polyslab(vertices, dilation, angle, bounds)
        # set up proper thickness
        _, max_dist = s._crossing_detection(s.base_polygon, -100)
        dilation = 0.0
        bounds = (0, max_dist * 0.99)
        angle = np.pi / 4
        # avoid vertex-edge crossing case
        try:
            s = setup_polyslab(vertices, dilation, angle, bounds)
        except:
            continue

        for axis in (0, 1):
            position = np.random.random(1)[0] * Lx - Lx / 2
            height = s._find_intersecting_height(position, axis)
            for h in height:
                bounds = (0, h)
                s = setup_polyslab(vertices, dilation, angle, bounds)
                diff = s.top_polygon[:, axis] - position
                assert np.any(np.isclose(diff, 0)) == True


def test_intersection_with_inside():
    """Make sure intersection produces the same result as inside"""

    N = 10  # number of vertices
    Lx = 10  # maximal length in x,y direction
    for i in range(50):
        vertices = convert_valid_polygon(np.random.random((N, 2)) * Lx)
        vertices = np.array(vertices)  # .astype("float32")
        dilation = 0
        angle = 0
        bounds = (0, 1)
        s = setup_polyslab(vertices, dilation, angle, bounds)

        # set up proper thickness
        _, max_dist = s._crossing_detection(s.base_polygon, -100)
        dilation = 0.0
        bounds = (0, (max_dist * 0.95))
        angle = np.pi / 4
        # avoid vertex-edge crossing case
        try:
            s = setup_polyslab(vertices, dilation, angle, bounds)
        except:
            continue

        ### side x
        xp = np.random.random(1)[0] * 2 * Lx - Lx
        yp = np.random.random(10) * 2 * Lx - Lx
        zp = np.random.random(10) * (bounds[1] - bounds[0]) + bounds[0]
        # xp = np.float32(xp)
        # yp = np.float32(yp)
        # zp = np.float32(zp)
        shape_intersect = s.intersections(x=xp)

        for i in range(len(yp)):
            for j in range(len(zp)):
                # inside
                res_inside = s.inside(xp, yp[i], zp[j])
                # intersect
                res_inter = False
                for shape in shape_intersect:
                    if shape.covers(Point(yp[i], zp[j])):
                        res_inter = True
                # if res_inter != res_inside:
                #     print(repr(vertices))
                #     print(repr(s.base_polygon))
                #     print(bounds)
                #     print(xp, yp[i], zp[j])
                assert res_inter == res_inside

        ### side y
        xp = np.random.random(10) * 2 * Lx - Lx
        yp = np.random.random(1)[0] * 2 * Lx - Lx
        zp = np.random.random(10) * (bounds[1] - bounds[0]) + bounds[0]
        # xp = np.float32(xp)
        # yp = np.float32(yp)
        # zp = np.float32(zp)
        shape_intersect = s.intersections(y=yp)

        for i in range(len(xp)):
            for j in range(len(zp)):
                # inside
                res_inside = s.inside(xp[i], yp, zp[j])
                # intersect
                res_inter = False
                for shape in shape_intersect:
                    if shape.covers(Point(xp[i], zp[j])):
                        res_inter = True
                assert res_inter == res_inside

        ### norm z
        xp = np.random.random(10) * 2 * Lx - Lx
        yp = np.random.random(10) * 2 * Lx - Lx
        zp = np.random.random(1)[0] * (bounds[1] - bounds[0]) + bounds[0]
        shape_intersect = s.intersections(z=zp)

        for i in range(len(xp)):
            for j in range(len(yp)):
                # inside
                res_inside = s.inside(xp[i], yp[j], zp)
                # intersect
                res_inter = False
                for shape in shape_intersect:
                    if shape.covers(Point(xp[i], yp[j])):
                        res_inter = True
                assert res_inter == res_inside


def test_bound():
    """
    Make sure bound works, even though it might not be tight.
    """
    N = 10  # number of vertices
    Lx = 10  # maximal length in x,y direction
    for i in range(50):
        vertices = convert_valid_polygon(np.random.random((N, 2)) * Lx)
        vertices = np.array(vertices)  # .astype("float32")

        ### positive dilation
        dilation = 0
        angle = 0
        bounds = (0, 1)
        s = setup_polyslab(vertices, dilation, angle, bounds)
        _, max_dist = s._crossing_detection(s.base_polygon, 100)
        # verify it is indeed maximal allowed
        dilation = 1
        if max_dist is not None:
            dilation = max_dist - 1e-10
        bounds = (0, 1)
        angle = 0.0
        # avoid vertex-edge crossing case
        try:
            s = setup_polyslab(vertices, dilation, angle, bounds)
        except:
            continue
        validate_poly_bound(s)

        ## sidewall
        dilation = 0
        angle = 0
        bounds = (0, 1)
        s = setup_polyslab(vertices, dilation, angle, bounds)
        # set up proper thickness
        _, max_dist = s._crossing_detection(s.base_polygon, -100)
        dilation = 0.0
        bounds = (0, (max_dist * 0.95))
        angle = np.pi / 4
        # avoid vertex-edge crossing case
        try:
            s = setup_polyslab(vertices, dilation, angle, bounds)
        except:
            continue
        validate_poly_bound(s)



"""Tools for dealing with static Simulation objects."""
import functools

from .log import Tidy3dError, log


class StaticException(Tidy3dError):
    """An exception when the tidy3d object has changed."""


class MakeStatic:
    """Context manager that stores a hash and checks if the object has changed upon teardown."""

    def __init__(self, simulation: "Simulation"):
        """When context manager is initalized, return a copy of the simulation."""
        if simulation.type != "Simulation":
            raise ValueError("simulation must be a ``Simulation`` type.")
        self.simulation = simulation.copy(deep=True)
        self.original_hash = self.simulation._freeze()

    def __enter__(self):
        """Freeze the simulation when entering context."""
        log.debug("-> entering static context")
        return self.simulation

    def __exit__(self, *args):
        """Unfeeze the simulation when leaving context, check the hashes equal."""
        log.debug("<- done with static context")
        final_hash = self.simulation._unfreeze()
        if final_hash != self.original_hash:
            raise StaticException("Simulation has changed in static context.")


def make_static(function):
    """Decorates a function to make the first argument (Simulation) static during the call."""

    @functools.wraps(function)
    def static_function(*args, **kwargs):

        # the first argument is assumed to be the simulation, note this can be customized later.
        sim_original, *args = args

        # call the original function within the static context manager
        with MakeStatic(sim_original) as sim_static:
            new_args = sim_static, *args
            ret_value = function(*new_args, **kwargs)

        return ret_value

    return static_function



"""Solve for modes in a 2D cross-sectional plane in a simulation, assuming translational
invariance along a given propagation axis.
"""

from typing import List, Tuple, Union, Dict
import logging

import h5py
import numpy as np
import pydantic

from ...components.base import Tidy3dBaseModel
from ...components import Box
from ...components import Simulation
from ...components import ModeSpec
from ...components import ModeMonitor
from ...components.source import ModeSource, SourceTime
from ...components.types import Direction, Array, Ax, Literal, ArrayLike, Axis, Symmetry
from ...components.data import Tidy3dData, ModeIndexData, ModeFieldData, ScalarModeFieldData
from ...components.data import AbstractSimulationData
from ...log import ValidationError, DataError

from .solver import compute_modes

FIELD = Tuple[Array[complex], Array[complex], Array[complex]]

# Warning for field intensity at edges over total field intensity larger than this value
FIELD_DECAY_CUTOFF = 1e-2


class ModeSolverData(AbstractSimulationData):
    """Holds data associated with :class:`.ModeSolver`.

    Parameters
    ----------
    plane : :class:`.Box`
        Cross-sectional plane in which the modes were be computed.
    mode_spec : :class:`.ModeSpec`
        Container with specifications about the modes.
    data_dict : Dict[str, Union[ModeFieldData, ModeIndexData]]
        Mapping of "n_complex" to :class:`.ModeIndexData`, and "fields" to :class:`.ModeFieldData`.
    """

    plane: Box
    mode_spec: ModeSpec
    data_dict: Dict[str, Union[ModeFieldData, ModeIndexData]]

    @property
    def fields(self):
        """Get field data."""
        return self.data_dict.get("fields")

    @property
    def n_complex(self):
        """Get complex effective indexes."""
        scalar_data = self.data_dict.get("n_complex")
        if scalar_data:
            return scalar_data.data
        return None

    @property
    def n_eff(self):
        """Get real part of effective index."""
        scalar_data = self.data_dict.get("n_complex")
        if scalar_data:
            return scalar_data.n_eff
        return None

    @property
    def k_eff(self):
        """Get imaginary part of effective index."""
        scalar_data = self.data_dict.get("n_complex")
        if scalar_data:
            return scalar_data.k_eff
        return None

    def add_to_handle(self, handle: Union[h5py.File, h5py.Group]) -> None:
        """Export to an hdf5 handle, which can be a file or a group.

        Parameters
        ----------
        handle : Union[hdf5.File, hdf5.Group]
            Handle to write the ModeSolverData to.
        """

        # save pydantic models as string
        json_dict = {
            "simulation": self.simulation,
            "plane": self.plane,
            "mode_spec": self.mode_spec,
        }
        for name, obj in json_dict.items():
            Tidy3dData.save_string(handle, name, obj.json())

        # make groups for mode fields and index data
        for name, data in self.data_dict.items():
            data_grp = handle.create_group(name)
            data.add_to_group(data_grp)

    @classmethod
    def load_from_handle(cls, handle: Union[h5py.File, h5py.Group]) -> "ModeSolverData":
        """Load from an hdf5 handle, which can be a file or a group.

        Parameters
        ----------
        handle : Union[hdf5.File, hdf5.Group]
            Handle to load the ModeSolverData from.
        """

        # construct pydantic models from string
        json_dict = {
            "simulation": Simulation,
            "plane": Box,
            "mode_spec": ModeSpec,
        }
        obj_dict = {}
        for name, obj in json_dict.items():
            json_string = Tidy3dData.load_string(handle, name)
            obj_dict[name] = obj.parse_raw(json_string)

        # load fields and effective index data
        data_dict = {
            "fields": ModeFieldData.load_from_group(handle["fields"]),
            "n_complex": ModeIndexData.load_from_group(handle["n_complex"]),
        }
        return cls(data_dict=data_dict, **obj_dict)

    def to_file(self, fname: str) -> None:
        """Export :class:`.ModeSolverData` to single hdf5 file.

        Parameters
        ----------
        fname : str
            Path to .hdf5 data file (including filename).
        """

        with h5py.File(fname, "a") as f_handle:
            self.add_to_handle(f_handle)

    @classmethod
    def from_file(cls, fname: str, **kwargs) -> "ModeSolverData":  # pylint:disable=unused-argument
        """Load :class:`.ModeSolverData` from .hdf5 file.

        Parameters
        ----------
        fname : str
            Path to .hdf5 data file (including filename).

        Returns
        -------
        :class:`.ModeSolverData`
            A :class:`.ModeSolverData` instance.
        """

        # read from file at fname
        with h5py.File(fname, "r") as f_handle:
            mode_data = cls.load_from_handle(f_handle)

        return mode_data

    # pylint:disable=too-many-arguments, too-many-locals, too-many-branches, too-many-statements
    def plot_field(
        self,
        field_name: str,
        val: Literal["real", "imag", "abs"] = "real",
        freq: float = None,
        mode_index: int = None,
        eps_alpha: float = 0.2,
        robust: bool = True,
        ax: Ax = None,
        **patch_kwargs,
    ) -> Ax:
        """Plot the field data for a monitor with simulation plot overlayed.

        Parameters
        ----------
        field_name : str
            Name of `field` to plot (eg. 'Ex').
            Also accepts `'int'` to plot intensity.
        val : Literal['real', 'imag', 'abs'] = 'real'
            Which part of the field to plot.
            If ``field_name='int'``, this has no effect.
        freq: float = None
            Specifies the frequency (Hz) to plot.
            Also sets the frequency at which the permittivity is evaluated at (if dispersive).
        mode_index: int = None
            Specifies which mode index to plot.
        eps_alpha : float = 0.2
            Opacity of the structure permittivity.
            Must be between 0 and 1 (inclusive).
        robust : bool = True
            If specified, uses the 2nd and 98th percentiles of the data to compute the color limits.
            This helps in visualizing the field patterns especially in the presence of a source.
        ax : matplotlib.axes._subplots.Axes = None
            matplotlib axes to plot on, if not specified, one is created.
        **patch_kwargs
            Optional keyword arguments passed to ``add_artist(patch, **patch_kwargs)``.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """

        if mode_index >= self.mode_spec.num_modes:
            raise DataError("``mode_index`` larger than ``mode_spec.num_modes``.")
        mode_fields = self.fields.sel_mode_index(mode_index=mode_index)

        # get the field data component
        if field_name == "int":
            xr_data = 0.0
            for field in ("Ex", "Ey", "Ez"):
                mode_fields = mode_fields[field]
                xr_data += abs(mode_fields) ** 2
            val = "abs"
        else:
            xr_data = mode_fields.data_dict.get(field_name).data

        field_data = xr_data.sel(f=freq, method="nearest")

        axis = self.plane.size.index(0.0)
        position = self.plane.center[axis]

        ax = self.plot_field_array(
            field_data=field_data,
            axis=axis,
            position=position,
            val=val,
            freq=freq,
            eps_alpha=eps_alpha,
            robust=robust,
            ax=ax,
            **patch_kwargs,
        )

        n_eff = self.n_eff.isel(mode_index=mode_index).sel(f=freq, method="nearest")
        title = f"f={float(field_data.f):1.2e}, n_eff={float(n_eff):1.4f}"
        ax.set_title(title)

        return ax


class ModeSolver(Tidy3dBaseModel):
    """Interface for solving electromagnetic eigenmodes in a 2D plane with translational
    invariance in the third dimension.
    """

    simulation: Simulation = pydantic.Field(
        ..., title="Simulation", description="Simulation defining all structures and mediums."
    )

    plane: Box = pydantic.Field(
        ..., title="Plane", description="Cross-sectional plane in which the mode will be computed."
    )

    mode_spec: ModeSpec = pydantic.Field(
        ...,
        title="Mode specification",
        description="Container with specifications about the modes to be solved for.",
    )

    freqs: Union[List[float], ArrayLike] = pydantic.Field(
        ..., title="Frequencies", description="A list of frequencies at which to solve."
    )

    @pydantic.validator("plane", always=True)
    def is_plane(cls, val):
        """Raise validation error if not planar."""
        if val.size.count(0.0) != 1:
            raise ValidationError(f"ModeSolver plane must be planar, given size={val}")
        return val

    @pydantic.validator("freqs", always=True)
    def freqs_not_empty(cls, val):
        """Raise validation error if ``freqs`` is an empty list."""
        if len(val) == 0:
            raise ValidationError("ModeSolver 'freqs' must be a non-empty list.")
        return val

    @property
    def normal_axis(self) -> Axis:
        """Axis normal to the mode plane."""
        return self.plane.size.index(0.0)

    @property
    def _plane_sym(self) -> Box:
        """Potentially smaller plane if symmetries present in the simulation."""
        return self.simulation.min_sym_box(self.plane)

    # pylint:disable=too-many-locals
    def solve(self) -> ModeSolverData:
        """Finds the modal profile and effective index of the modes.

        Returns
        -------
        ModeSolverData
            :class:`.ModeSolverData` object containing the effective index and mode fields for all
            modes.
        """

        normal_axis = self.normal_axis

        # get the in-plane grid coordinates on which eps and the mode fields live
        plane_grid = self.simulation.discretize(self.plane)

        # restrict to a smaller plane if symmetries present in the simulation
        plane_grid_sym = self.simulation.discretize(self._plane_sym)

        # Coords and symmetry arguments to the solver (restricted to in-plane)
        _, solver_coords = self.plane.pop_axis(plane_grid_sym.boundaries.to_list, axis=normal_axis)
        mode_symmetry = list(self.simulation.symmetry)
        for dim in range(3):
            if self.simulation.center[dim] != self.plane.center[dim]:
                mode_symmetry[dim] = 0
        _, solver_symmetry = self.plane.pop_axis(mode_symmetry, axis=normal_axis)

        # Compute and store the modes at all frequencies
        n_complex, fields = self._solve_all_freqs(coords=solver_coords, symmetry=solver_symmetry)

        # Generate the dictionary of ScalarModeFieldData for every field
        data_dict = {}
        for field_name in fields[0].keys():
            xyz_coords = plane_grid_sym[field_name].to_list
            xyz_coords[normal_axis] = [self.plane.center[normal_axis]]
            data_dict[field_name] = ScalarModeFieldData(
                x=xyz_coords[0],
                y=xyz_coords[1],
                z=xyz_coords[2],
                f=self.freqs,
                mode_index=np.arange(self.mode_spec.num_modes),
                values=np.stack([field_freq[field_name] for field_freq in fields], axis=-2),
            )
        field_data = ModeFieldData(
            data_dict=data_dict,
            expanded_grid=plane_grid.yee.grid_dict,
            symmetry_center=self.simulation.center,
            symmetry=self.simulation.symmetry,
        )
        field_data = field_data.expand_syms
        self._field_decay_warning(field_data)
        index_data = ModeIndexData(
            f=self.freqs,
            mode_index=np.arange(self.mode_spec.num_modes),
            values=np.stack(n_complex, axis=0),
        )
        mode_data = ModeSolverData(
            simulation=self.simulation,
            plane=self.plane,
            mode_spec=self.mode_spec,
            data_dict={"fields": field_data, "n_complex": index_data},
        )

        return mode_data

    def _get_epsilon(self, plane: Box, freq: float) -> Array[complex]:
        """Compute the diagonal components of the epsilon tensor in the plane."""

        eps_xx = self.simulation.epsilon(plane, "Ex", freq)
        eps_yy = self.simulation.epsilon(plane, "Ey", freq)
        eps_zz = self.simulation.epsilon(plane, "Ez", freq)

        return np.stack((eps_xx, eps_yy, eps_zz), axis=0)

    def _solver_eps(self, freq: float) -> Array[complex]:
        """Get the diagonal permittivity in the shape needed to be supplied to the sovler, with the
        normal axis rotated to z."""

        # Get diagonal epsilon components in the plane
        (eps_xx, eps_yy, eps_zz) = self._get_epsilon(self._plane_sym, freq)

        # get rid of normal axis
        eps_xx = np.squeeze(eps_xx, axis=self.normal_axis)
        eps_yy = np.squeeze(eps_yy, axis=self.normal_axis)
        eps_zz = np.squeeze(eps_zz, axis=self.normal_axis)

        # swap axes to plane coordinates (normal_axis goes to z)
        eps_sim_ax = (eps_xx, eps_yy, eps_zz)
        eps_zz, (eps_xx, eps_yy) = self.plane.pop_axis(eps_sim_ax, axis=self.normal_axis)

        # construct eps to feed to mode solver
        return np.stack((eps_xx, eps_yy, eps_zz), axis=0)

    def _solve_all_freqs(
        self, coords: Tuple[Array, Array], symmetry: Tuple[Symmetry, Symmetry]
    ) -> Tuple[List[float], List[Dict[str, Array[complex]]]]:
        """Call the mode solver at all requested frequencies."""

        fields = []
        n_complex = []
        for freq in self.freqs:
            n_freq, fields_freq = self._solve_single_freq(
                freq=freq, coords=coords, symmetry=symmetry
            )
            fields.append(fields_freq)
            n_complex.append(n_freq)

        return n_complex, fields

    def _solve_single_freq(
        self, freq: float, coords: Tuple[Array, Array], symmetry: Tuple[Symmetry, Symmetry]
    ) -> Tuple[float, Dict[str, Array[complex]]]:
        """Call the mode solver at a single frequency.
        The fields are rotated from propagation coordinates back to global coordinates.
        """

        solver_fields, n_complex = compute_modes(
            eps_cross=self._solver_eps(freq),
            coords=coords,
            freq=freq,
            mode_spec=self.mode_spec,
            symmetry=symmetry,
        )

        fields = {"Ex": [], "Ey": [], "Ez": [], "Hx": [], "Hy": [], "Hz": []}
        for mode_index in range(self.mode_spec.num_modes):
            # Get E and H fields at the current mode_index
            ((Ex, Ey, Ez), (Hx, Hy, Hz)) = self._process_fields(solver_fields, mode_index)

            # Note: back in original coordinates
            fields_mode = {"Ex": Ex, "Ey": Ey, "Ez": Ez, "Hx": Hx, "Hy": Hy, "Hz": Hz}
            for field_name, field in fields_mode.items():
                fields[field_name].append(field)

        for field_name, field in fields.items():
            fields[field_name] = np.stack(field, axis=-1)

        return n_complex, fields

    def _rotate_field_coords(self, field: FIELD) -> FIELD:
        """Move the propagation axis=z to the proper order in the array."""
        f_x, f_y, f_z = np.moveaxis(field, source=3, destination=1 + self.normal_axis)
        f_rot = np.stack(self.plane.unpop_axis(f_z, (f_x, f_y), axis=self.normal_axis), axis=0)
        return f_rot

    def _process_fields(self, mode_fields: Array[complex], mode_index: int) -> Tuple[FIELD, FIELD]:
        """Transform solver fields to simulation axes, set gauge, and check decay at boundaries."""

        # Separate E and H fields (in solver coordinates)
        E, H = mode_fields[..., mode_index]

        # Set gauge to highest-amplitude in-plane E being real and positive
        ind_max = np.argmax(np.abs(E[:2]))
        phi = np.angle(E[:2].ravel()[ind_max])
        E *= np.exp(-1j * phi)
        H *= np.exp(-1j * phi)

        # Rotate back to original coordinates
        (Ex, Ey, Ez) = self._rotate_field_coords(E)
        (Hx, Hy, Hz) = self._rotate_field_coords(H)

        # apply -1 to H fields if a reflection was involved in the rotation
        if self.normal_axis == 1:
            Hx *= -1
            Hy *= -1
            Hz *= -1

        return ((Ex, Ey, Ez), (Hx, Hy, Hz))

    def _field_decay_warning(self, field_data: ModeFieldData):
        """Warn if any of the modes do not decay at the edges."""
        _, plane_dims = self.plane.pop_axis(["x", "y", "z"], axis=self.normal_axis)
        field_sizes = field_data.Ex.sizes
        for freq_index in range(field_sizes["f"]):
            for mode_index in range(field_sizes["mode_index"]):
                e_edge, e_norm = 0, 0
                # Sum up the total field intensity
                for E in (field_data.Ex, field_data.Ey, field_data.Ez):
                    e_norm += np.sum(np.abs(E[{"f": freq_index, "mode_index": mode_index}]) ** 2)
                # Sum up the field intensity at the edges
                if field_sizes[plane_dims[0]] > 1:
                    for E in (field_data.Ex, field_data.Ey, field_data.Ez):
                        isel = {plane_dims[0]: [0, -1], "f": freq_index, "mode_index": mode_index}
                        e_edge += np.sum(np.abs(E[isel]) ** 2)
                if field_sizes[plane_dims[1]] > 1:
                    for E in (field_data.Ex, field_data.Ey, field_data.Ez):
                        isel = {plane_dims[1]: [0, -1], "f": freq_index, "mode_index": mode_index}
                        e_edge += np.sum(np.abs(E[isel]) ** 2)
                # Warn if needed
                if e_edge / e_norm > FIELD_DECAY_CUTOFF:
                    logging.warning(
                        f"Mode field at frequency index {freq_index}, mode index {mode_index} does "
                        "not decay at the plane boundaries."
                    )

    def to_source(
        self,
        source_time: SourceTime,
        direction: Direction,
        mode_index: int = 0,
    ) -> ModeSource:
        """Creates :class:`.ModeSource` from a :class:`.ModeSolver` instance plus additional
        specifications.

        Parameters
        ----------
        source_time: :class:`.SourceTime`
            Specification of the source time-dependence.
        direction : Direction
            Whether source will inject in ``"+"`` or ``"-"`` direction relative to plane normal.
        mode_index : int = 0
            Index into the list of modes returned by mode solver to use in source.

        Returns
        -------
        :class:`.ModeSource`
            Mode source with specifications taken from the ModeSolver instance and the method
            inputs.
        """

        return ModeSource(
            center=self.plane.center,
            size=self.plane.size,
            source_time=source_time,
            mode_spec=self.mode_spec,
            mode_index=mode_index,
            direction=direction,
        )

    def to_monitor(self, freqs: List[float], name: str) -> ModeMonitor:
        """Creates :class:`ModeMonitor` from a :class:`.ModeSolver` instance plus additional
        specifications.

        Parameters
        ----------
        freqs : List[float]
            Frequencies to include in Monitor (Hz).
        name : str
            Required name of monitor.

        Returns
        -------
        :class:`.ModeMonitor`
            Mode monitor with specifications taken from the ModeSolver instance and the method
            inputs.
        """

        return ModeMonitor(
            center=self.plane.center,
            size=self.plane.size,
            freqs=freqs,
            mode_spec=self.mode_spec,
            name=name,
        )



"""Defines specification for mode solver."""

from typing import Tuple

import pydantic as pd
import numpy as np

from ..constants import MICROMETER, RADIAN, GLANCING_CUTOFF
from .base import Tidy3dBaseModel
from .types import Axis2D, Literal
from ..log import SetupError


class ModeSpec(Tidy3dBaseModel):
    """Stores specifications for the mode solver to find an electromagntic mode.
    Note, the planar axes are found by popping the injection axis from {x,y,z}.
    For example, if injection axis is y, the planar axes are ordered {x,z}.

    Example
    -------
    >>> mode_spec = ModeSpec(num_modes=3, target_neff=1.5)
    """

    num_modes: pd.PositiveInt = pd.Field(
        1, title="Number of modes", description="Number of modes returned by mode solver."
    )

    target_neff: pd.PositiveFloat = pd.Field(
        None, title="Target effective index", description="Guess for effective index of the mode."
    )

    num_pml: Tuple[pd.NonNegativeInt, pd.NonNegativeInt] = pd.Field(
        (0, 0),
        title="Number of PML layers",
        description="Number of standard pml layers to add in the two tangential axes.",
    )

    sort_by: Literal["largest_neff", "te_fraction", "tm_fraction"] = pd.Field(
        "largest_neff",
        title="Ordering of the returned modes",
        description="The solver will always compute the ``num_modes`` modes closest to the "
        "``target_neff``, but they can be reordered by the largest ``te_fraction``, defined "
        "as the integral of the intensity of the E-field component parallel to the first plane "
        "axis normalized to the total in-plane E-field intensity. Similarly, ``tm_fraction`` "
        "uses the E field component parallel to the second plane axis.",
    )

    angle_theta: float = pd.Field(
        0.0,
        title="Polar Angle",
        description="Polar angle of the propagation axis from the injection axis.",
        units=RADIAN,
    )

    angle_phi: float = pd.Field(
        0.0,
        title="Azimuth Angle",
        description="Azimuth angle of the propagation axis in the plane orthogonal to the "
        "injection axis.",
        units=RADIAN,
    )

    bend_radius: float = pd.Field(
        None,
        title="Bend radius",
        description="A curvature radius for simulation of waveguide bends. Can be negative, in "
        "which case the mode plane center has a smaller value than the curvature center along the "
        "tangential axis perpendicular to the bend axis.",
        units=MICROMETER,
    )

    bend_axis: Axis2D = pd.Field(
        None,
        title="Bend axis",
        description="Index into the two tangential axes defining the normal to the "
        "plane in which the bend lies. This must be provided if ``bend_radius`` is not ``None``. "
        "For example, for a ring in the global xy-plane, and a mode plane in either the xz or the "
        "yz plane, the ``bend_axis`` is always 1 (the global z axis).",
    )

    @pd.validator("bend_axis", always=True)
    def bend_axis_given(cls, val, values):
        """check that ``bend_axis`` is provided if ``bend_radius`` is not ``None``"""
        if val is None and values.get("bend_radius") is not None:
            raise SetupError("bend_axis must also be defined if bend_radius is defined.")
        return val

    @pd.validator("angle_theta", allow_reuse=True, always=True)
    def glancing_incidence(cls, val):
        """Warn if close to glancing incidence."""
        if np.abs(np.pi / 2 - val) < GLANCING_CUTOFF:
            raise SetupError(
                "Mode propagation axis too close to glancing angle for accurate injection. "
                "For best results, switch the injection axis."
            )
        return val



"""Tools for generating an S matrix automatically from tidy3d simulation and port definitions."""

from typing import List, Tuple, Dict

import pydantic as pd
import numpy as np

from ...constants import HERTZ, C_0
from ...components.simulation import Simulation
from ...components.geometry import Box
from ...components.mode import ModeSpec
from ...components.monitor import ModeMonitor
from ...components.source import ModeSource, GaussianPulse
from ...components.data import SimulationData
from ...components.types import Direction, Ax
from ...components.viz import add_ax_if_none, equal_aspect
from ...components.base import Tidy3dBaseModel
from ...log import SetupError, log
from ...web.container import Batch

# fwidth of gaussian pulse in units of central frequency
FWIDTH_FRAC = 1.0 / 10
DEFAULT_DATA_DIR = "data"


class Port(Box):
    """Specifies a port in the scattering matrix."""

    direction: Direction = pd.Field(
        ...,
        title="Direction",
        description="'+' or '-', defining which direction is considered 'input'.",
    )
    mode_spec: ModeSpec = pd.Field(
        ModeSpec(),
        title="Mode Specification",
        description="Specifies how the mode solver will solve for the modes of the port.",
    )
    mode_indices: Tuple[pd.NonNegativeInt] = pd.Field(
        None,
        title="Mode Indices.",
        description="Indices into modes returned by the mode solver to use in the port. "
        "If ``None``, all modes returned by the mode solver are used in the scattering matrix. "
        "Otherwise, the scattering matrix will include elements for each supplied mode index.",
    )
    name: str = pd.Field(
        ...,
        title="Name",
        description="Unique name for the port.",
        min_length=1,
    )

    @pd.validator("mode_indices", always=True)
    def evaluate_mode_indices(cls, val, values):
        """Evaluates mode indices based on number of modes in mode spec."""
        if val is None:
            num_modes = values.get("mode_spec").num_modes
            val = tuple(range(num_modes))
        return val


"""
s_matrix[port_name_in][port_name_out] gives a numpy array of shape (m, n)
relating the coupling amplitudes between the m and n mode orders in the two ports, respectively.
"""
SMatrixType = Dict[str, np.ndarray]


class ComponentModeler(Tidy3dBaseModel):
    """Tool for modeling devices and computing scattering matrix elements."""

    simulation: Simulation = pd.Field(
        ...,
        title="Simulation",
        description="Simulation describing the device without any sources or monitors present.",
    )
    ports: List[Port] = pd.Field(
        [],
        title="Ports",
        description="Collection of ports describing the scattering matrix elements. "
        "For each port, one simulation will be run with a modal source.",
    )
    freq: float = pd.Field(
        ...,
        title="Frequency",
        description="Frequency at which to evaluate the scattering matrix.",
        units=HERTZ,
    )

    batch_data: Tidy3dBaseModel = pd.Field(
        None,
        title="Batch Data",
        description="Batch Data of task used to compute S matrix. Set internally.",
    )

    @pd.validator("simulation", always=True)
    def _sim_has_no_sources(cls, val):
        """Make sure simulation has no sources as they interfere with tool."""
        if len(val.sources) > 0:
            raise SetupError("Simulation must not have sources.")
        return val

    def _to_monitor(self, port: Port) -> ModeMonitor:
        """Creates a mode monitor from a given port."""
        return ModeMonitor(
            center=port.center,
            size=port.size,
            freqs=[self.freq],
            mode_spec=port.mode_spec,
            name=port.name,
        )

    def _to_sources(self, port: Port) -> List[ModeSource]:
        """Creates a list of mode sources from a given port."""
        return [
            ModeSource(
                center=port.center,
                size=port.size,
                source_time=GaussianPulse(freq0=self.freq, fwidth=self.freq * FWIDTH_FRAC),
                mode_spec=port.mode_spec,
                mode_index=mode_index,
                direction=port.direction,
                name=port.name,
            )
            for mode_index in port.mode_indices
        ]

    @equal_aspect
    @add_ax_if_none
    def plot_sim(self, x: float = None, y: float = None, z: float = None, ax: Ax = None) -> Ax:
        """Plot a :class:`Simulation` with all sources added for each port, for troubleshooting."""

        sim_plot = self.simulation.copy(deep=True)
        for port_source in self.ports:
            mode_source_0 = self._to_sources(port_source)[0]
            sim_plot.sources.append(mode_source_0)
        return sim_plot.plot(x=x, y=y, z=z, ax=ax)

    def _shift_value_signed(self, port: Port) -> float:
        """How far (signed) to shift the source from the monitor."""

        # get the grid boundaries and sizes along port normal from the simulation
        normal_axis = port.size.index(0.0)
        grid = self.simulation.grid
        grid_boundaries = grid.boundaries.to_list[normal_axis]
        grid_centers = grid.centers.to_list[normal_axis]

        # get the index of the grid cell where the port lies
        port_position = port.center[normal_axis]
        port_index = np.argwhere(port_position > grid_boundaries)[-1]

        # shift the port to the left
        if port.direction == "+":
            shifted_index = port_index - 2
            if shifted_index < 0:
                raise SetupError(
                    f"Port {port.name} normal is too close to boundary "
                    f"on -{'xyz'[normal_axis]} side."
                )

        # shift the port to the right
        else:
            shifted_index = port_index + 2
            if shifted_index >= len(grid_centers):
                raise SetupError(
                    f"Port {port.name} normal is too close to boundary "
                    f"on +{'xyz'[normal_axis]} side."
                )

        new_pos = grid_centers[shifted_index]
        return new_pos - port_position

    def _shift_port(self, port: Port) -> Port:
        """Generate a new port shifted by the shift amount in normal direction."""

        shift_value = self._shift_value_signed(port)
        center_shifted = list(port.center)
        center_shifted[port.size.index(0.0)] += shift_value
        port_shifted = port.copy(deep=True)
        port_shifted.center = center_shifted
        return port_shifted

    def _task_name(self, port_source: Port, mode_index: int) -> str:
        """The name of a task, determined by the port of the source and mode index."""
        return f"smatrix_port{port_source.name}_mode{mode_index}"

    def _make_sims(self) -> Dict[str, Simulation]:
        """Generate all the :class:`Simulation` objects for the S matrix calculation."""

        mode_monitors = [self._to_monitor(port) for port in self.ports]
        sim_dict = {}
        for port_source in self.ports:
            port_source = self._shift_port(port_source)
            for mode_source in self._to_sources(port_source):
                sim_copy = self.simulation.copy(deep=True)
                sim_copy.sources = [mode_source]
                sim_copy.monitors += mode_monitors
                task_name = self._task_name(port_source, mode_source.mode_index)
                sim_dict[task_name] = sim_copy
        return sim_dict

    def _run_sims(
        self, sim_dict: Dict[str, Simulation], folder_name: str, path_dir: str
    ) -> "BatchData":
        """Run :class:`Simulations` for each port and return the batch after saving."""

        batch = Batch(simulations=sim_dict, folder_name=folder_name)
        batch.upload()
        batch.start()
        batch.monitor()
        batch_data = batch.load(path_dir=path_dir)
        self.batch_data = batch_data
        return batch_data

    def _normalization_factor(self, port_source: Port, sim_data: SimulationData) -> complex:
        """Compute the normalization amplitude based on the measured input mode amplitude."""

        port_monitor_data = sim_data[port_source.name]
        mode_index = sim_data.simulation.sources[0].mode_index

        normalize_amp = port_monitor_data.amps.sel(
            f=self.freq,
            direction=port_source.direction,
            mode_index=mode_index,
        ).values

        normalize_n_eff = port_monitor_data.n_eff.sel(f=self.freq, mode_index=mode_index).values

        k0 = 2 * np.pi * C_0 / self.freq
        k_eff = k0 * normalize_n_eff
        shift_value = self._shift_value_signed(port_source)
        return normalize_amp * np.exp(1j * k_eff * shift_value)

    def _construct_smatrix(self, batch_data: "BatchData") -> SMatrixType:
        """Post process batch to generate scattering matrix."""

        # load all data
        s_matrix_dict = {
            port_source.name: {port_monitor.name: [] for port_monitor in self.ports}
            for port_source in self.ports
        }

        # loop through source ports
        for port_source in self.ports:

            source_name = port_source.name

            # loop through all source injection indices
            for mode_index in port_source.mode_indices:

                # get the data for this source and compute normalization by injection
                task_name = self._task_name(port_source, mode_index)
                sim_data = batch_data[task_name]
                norm_factor = self._normalization_factor(port_source, sim_data)

                # loop through all monitors
                for port_monitor in self.ports:

                    monitor_name = port_monitor.name

                    # compute the mode amplitude data
                    dir_out = "-" if port_monitor.direction == "+" else "+"
                    mode_data = sim_data[monitor_name].amps.sel(f=self.freq, direction=dir_out)
                    amps_normalized = np.array(mode_data.values) / norm_factor
                    s_matrix_dict[source_name][monitor_name].append(amps_normalized)

            # convert to an array
            for port_monitor in self.ports:
                monitor_name = port_monitor.name
                mode_matrix = np.array(s_matrix_dict[source_name][monitor_name])
                s_matrix_dict[source_name][monitor_name] = mode_matrix

        return s_matrix_dict

    def solve(self, folder_name: str = "default", path_dir: str = DEFAULT_DATA_DIR) -> SMatrixType:
        """Solves for the scattering matrix of the system."""
        log.warning(
            "`ComponentModeler.solve()` is renamed to `ComponentModeler.run()` "
            "'and will be removed in a later version."
        )
        return self.run(folder_name=folder_name, path_dir=path_dir)

    def run(self, folder_name: str = "default", path_dir: str = DEFAULT_DATA_DIR) -> SMatrixType:
        """Solves for the scattering matrix of the system."""

        sim_dict = self._make_sims()
        batch_data = self._run_sims(sim_dict=sim_dict, folder_name=folder_name, path_dir=path_dir)
        return self._construct_smatrix(batch_data=batch_data)

    def load(self) -> SMatrixType:
        """Load an Smatrix from saved BatchData object."""

        if self.batch_data is None:
            raise SetupError("Component modeler has no batch saved. Run .solve() to generate.")
        return self._construct_smatrix(batch_data=self.batch_data)



# pylint: disable=invalid-name
""" utilities for plotting """
from typing import Any
from functools import wraps

import matplotlib.pylab as plt
import pydantic as pd

from .types import Ax
from .base import Tidy3dBaseModel

""" Constants """

# add this around extents of plots
PLOT_BUFFER = 0.3

ARROW_COLOR_MONITOR = "orange"
ARROW_COLOR_SOURCE = "green"
ARROW_COLOR_POLARIZATION = "brown"
ARROW_ALPHA = 0.8


# this times the min of axis height and width gives the arrow length
ARROW_LENGTH_FACTOR = 0.4

# this times ARROW_LENGTH gives width
ARROW_WIDTH_FACTOR = 0.4

# arrow width cannot be larger than this factor times the max of axis height and width
MAX_ARROW_WIDTH_FACTOR = 0.02


""" Decorators """


def make_ax() -> Ax:
    """makes an empty `ax`."""
    _, ax = plt.subplots(1, 1, tight_layout=True)
    return ax


def add_ax_if_none(plot):
    """Decorates `plot(*args, **kwargs, ax=None)` function.
    if ax=None in the function call, creates an ax and feeds it to rest of function.
    """

    @wraps(plot)
    def _plot(*args, **kwargs) -> Ax:
        """New plot function using a generated ax if None."""
        if kwargs.get("ax") is None:
            ax = make_ax()
            kwargs["ax"] = ax
        return plot(*args, **kwargs)

    return _plot


def equal_aspect(plot):
    """Decorates a plotting function returning a matplotlib axes.
    Ensures the aspect ratio of the returned axes is set to equal.
    Useful for 2D plots, like sim.plot() or sim_data.plot_fields()
    """

    @wraps(plot)
    def _plot(*args, **kwargs) -> Ax:
        """New plot function with equal aspect ratio axes returned."""
        ax = plot(*args, **kwargs)
        ax.set_aspect("equal")
        return ax

    return _plot


""" plot parameters """


class PlotParams(Tidy3dBaseModel):
    """Stores plotting parameters / specifications for a given model."""

    alpha: Any = pd.Field(1.0, title="Opacity")
    edgecolor: Any = pd.Field(None, title="Edge Color", alias="ec")
    facecolor: Any = pd.Field(None, title="Face Color", alias="fc")
    fill: bool = pd.Field(True, title="Is Filled")
    hatch: str = pd.Field(None, title="Hatch Style")
    linewidth: pd.NonNegativeFloat = pd.Field(1, title="Line Width", alias="lw")

    def include_kwargs(self, **kwargs) -> "PlotParams":
        """Update the plot params with supplied kwargs."""
        new_plot_params = self.copy(deep=True)
        for key, value in kwargs.items():
            if key not in ("type",) and value is not None:
                setattr(new_plot_params, key, value)
        return new_plot_params

    def to_kwargs(self) -> dict:
        """Export the plot parameters as kwargs dict that can be supplied to plot function."""
        kwarg_dict = self.dict()
        for ignore_key in ("type",):
            kwarg_dict.pop(ignore_key)
        return kwarg_dict


# defaults for different tidy3d objects
plot_params_geometry = PlotParams()
plot_params_structure = PlotParams()
plot_params_source = PlotParams(alpha=0.4, facecolor="limegreen", edgecolor="limegreen", lw=3)
plot_params_monitor = PlotParams(alpha=0.4, facecolor="orange", edgecolor="orange", lw=3)
plot_params_pml = PlotParams(alpha=0.7, facecolor="gray", edgecolor="gray", hatch="x")
plot_params_symmetry = PlotParams(edgecolor="gray", facecolor="gray", alpha=0.6)
plot_params_override_structures = PlotParams(linewidth=0.4, edgecolor="black", fill=False)

# stores color of simulation.structures for given index in simulation.medium_map
MEDIUM_CMAP = [
    "#689DBC",
    "#D0698E",
    "#5E6EAD",
    "#C6224E",
    "#BDB3E2",
    "#9EC3E0",
    "#616161",
    "#877EBC",
]
